#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2023 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ruff: noqa: SLF001

"""
Unit tests for queue with limited number of items.
"""

import asyncio
from n23.queue import NQueue

import pytest

def test_queue_put() -> None:
    """
    Test adding items to queue.
    """
    queue = NQueue[int](3)
    queue.put(1)
    queue.put(2)
    queue.put(3)
    queue.put(4)

    assert list(queue._data) == [2, 3, 4]

@pytest.mark.timeout(3)
@pytest.mark.asyncio
async def test_queue_get() -> None:
    """
    Test getting items from queue.
    """
    queue = NQueue[int](3)
    async def add() -> bool:
        await asyncio.sleep(0.2)
        queue.put(101)
        return True

    mark_add, result = await asyncio.gather(add(), queue.get())
    assert mark_add is True
    assert result == 101

@pytest.mark.timeout(3)
@pytest.mark.asyncio
async def test_queue_get_twice() -> None:
    """
    Test type error for two consumer of a queue.
    """
    queue = NQueue[int](3)
    with pytest.raises(TypeError):
        await asyncio.gather(queue.get(), queue.get())

@pytest.mark.timeout(3)
@pytest.mark.asyncio
async def test_queue_size() -> None:
    queue = NQueue[int](3)
    queue.put(1)
    queue.put(2)
    assert not queue.full()
    assert len(queue) == 2

    queue.put(3)
    assert queue.full()
    assert len(queue) == 3

    queue.put(4)
    assert len(queue) == 3
    assert list(queue._data) == [2, 3, 4]

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2023 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Helper functions to start applications based on N23 framework.
"""

import asyncio
import logging
import signal
import sys
import typing as tp
import uvloop
from collections.abc import Coroutine
from unittest import mock

logger = logging.getLogger(__name__)

T = tp.TypeVar('T')
S = tp.TypeVar('S')
V = tp.TypeVar('V')

def run(coro: Coroutine[T, S, V]) -> V:
    """
    Run application based on N23 framework.

    The function provides the following benefits

    - uvloop based asyncio loop
    - TERM and INT signal handlers are installed to allow graceful shutdown
      of application components
    """
    uvloop.install()
    logger.info('uvloop installed')
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    with mock.patch.object(asyncio.events, 'new_event_loop') as mock_new_loop:
        mock_new_loop.return_value = loop

        # shutdown gracefully on SIGTERM and SIGINT
        loop.add_signal_handler(signal.SIGTERM, sys.exit)
        loop.add_signal_handler(signal.SIGINT, sys.exit)

        return asyncio.run(coro)

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2024 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Collection of functions to support functional programming in Python.
"""

import typing as tp
from collections.abc import Iterable

from n23.fn import head, tail, compose

import pytest

T = tp.TypeVar('T')

@pytest.mark.parametrize(
    "values, expected",
    [[[1, 2, 3], 1],
     [range(2, 5), 2]]
)
def test_head(values: Iterable[T], expected: T) -> None:
    """
    Test getting head item of an iterable.
    """
    assert head(values) == expected

@pytest.mark.parametrize(
    "values, expected",
    [[[1, 2, 3], [2, 3]],
     [range(2, 5), [3, 4]]]
)
def test_tail(values: Iterable[T], expected: list[T]) -> None:
    """
    Test getting head item of an iterable.
    """
    assert list(tail(values)) == expected

def test_compose_0() -> None:
    """
    Test function composition when the first function has no arguments.
    """
    result = compose(lambda: 1, lambda v: v * 3)()  # type: ignore
    assert result == 3

def test_compose() -> None:
    """
    Test function composition.
    """
    result = compose(lambda v: v + 1, lambda v: v * 3, lambda v: v - 1)(2)  # type: ignore
    assert result == 8

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2025 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
N23 scheduler to run asynchronous tasks processing data and executing
actions.

There are three types of tasks

- data task executed at regular time intervals; data received by the task
  is transformed by a pipeline function, and then by a sink asynchronous
  coroutine
- event-driven data tasks; data is received as it arrives, and is processed
  like above
- action tasks executed at regular time intervals; no data is received

The data tasks execution is split between 3 stages

- receive data with an asynchronous coroutine
- transform data with a pipeline function, which is a regular Python
  callable
- send data to a sink, which is an asynchronous coroutine

Different kinds of Python functions can be actions or receive data. These
are coroutines, asynchronous generators, awaitables, and regular Python
functions. They are always used as an asynchronous coroutine, i.e.
a regular function is run with `asyncio.to_thread`.

Design Notes
============
1. Task execution is separated from processing of a data of the task. This
   allows enqueue the data if processing happens to be too slow

   - the oldest data is lost when the queue is full
   - in the future, let's block a task when its data queue is full with
     a task policy

2. Tasks scheduled at regular intervals have timeout equal to 75% of time
   interval to receive data. At the next edge of the interval, non-pending
   tasks are rescheduled, and pending tasks are awaited

   - in the future the regular tasks timeout property needs to be
     configurable
   - in the future, allow to cancel tasks instead of waiting for them with
     a task policy

3. `atimer <https://wrobell.dcmod.org/atimer/>`_ library is used for
   scheduler timer. This allows expiration at regular intervals and
   reporting of timer overruns.

4. The data tasks execution is split between 3 stages to emphasize the
   rule, that data should not be received or sent during data
   transformation.

5. The functions receiving data shall not block the scheduler, therefore
   regular Python functions are turned into coroutines with
   `asyncio.to_thread` function.
"""

from __future__ import annotations

import asyncio
import enum
import inspect
import logging
import time
import typing as tp
from collections.abc import Awaitable, Coroutine, AsyncGenerator, \
    AsyncIterator, Generator, Iterable
from dataclasses import dataclass
from itertools import islice

from .fn import concat, partial
from .timer import Timer
from .queue import NQueue

logger = logging.getLogger(__name__)

type TaskCoroutine[T] = Coroutine[None, None, T]
type TaskCoroutineFunction[T] = tp.Callable[[], TaskCoroutine[T]]
type SourceFunction[T] = Awaitable[T] \
    | tp.Callable[
        [],
        T \
        | TaskCoroutine[T] \
        | AsyncGenerator[T] \
        | AsyncIterator[T]
]
type PipelineFunction[T, S] = tp.Callable[['TaskResult[T]'], S | None]
type SinkFunction[T] = tp.Callable[[T], Awaitable[tp.Any]]
type ActionFunction = SourceFunction[tp.Any | None]
type DataTask = 'RegularTask[tp.Any, tp.Any]' | 'EventDrivenTask[tp.Any, tp.Any]'

@dataclass(frozen=True, slots=True)
class TaskHandle[T]:
    """
    Scheduler task handle.
    """
    name: str

@dataclass(frozen=True, slots=True)
class TaskResult[T]:
    """
    Result of task execution.
    """
    time: float
    name: str
    value: T

class TimeoutPolicy(enum.Enum):
    """
    Timeout policies for N23 scheduler tasks.
    """
    CONTINUE = enum.auto()

class BufferPolicy(enum.Enum):
    """
    Buffer policies for N23 scheduler tasks.
    """
    DROP_OLD = enum.auto()

@dataclass(frozen=True, slots=True)
class TaskPolicy:
    """
    N23 scheduler task policy.

    :var on_timeout: What to do on scheduler task timeout.
    :var on_buffer_full: What to do when scheduler task buffer is full.
    :var buffer_limit: N23 scheduler task buffer limit.
    """
    on_timeout: TimeoutPolicy=TimeoutPolicy.CONTINUE
    on_buffer_full: BufferPolicy=BufferPolicy.DROP_OLD
    buffer_limit: int=1000

DEFAULT_TASK_POLICY = TaskPolicy()

@dataclass(frozen=False, slots=True)
class Context:
    """
    N23 scheduler context.

    :var tick: Timer cycle tick.
    :var time: Time at the start of timer cycle.
    """
    tick: int
    time: float

@dataclass(frozen=True, slots=True)
class Task[T]:
    handle: TaskHandle[T]

@dataclass(frozen=True, slots=True)
class RegularTask[T, S](Task[T]):
    interval: float
    source: SourceFunction[T]
    pipeline: PipelineFunction[T, S]
    sink: SinkFunction[S]
    queue: NQueue[TaskResult[T]]

@dataclass(frozen=True, slots=True)
class EventDrivenTask[T, S](Task[T]):
    source: SourceFunction[T]
    pipeline: PipelineFunction[T, S]
    sink: SinkFunction[S]
    queue: NQueue[TaskResult[T]]

@dataclass(frozen=True, slots=True)
class ActionTask(Task[tp.Any]):
    interval: float
    action: ActionFunction

class Scheduler(Awaitable[None]):
    """
    N23 scheduler to run asynchronous tasks processing data and executing
    actions.

    :var ctx: Scheduler context.
    :var timeout: Asynchronous task timeout.
    """
    def __init__(self, *, timeout: float | None=None):
        self.timeout = timeout

        self._r_tasks: list[RegularTask[tp.Any, tp.Any]] = []
        self._e_tasks: list[EventDrivenTask[tp.Any, tp.Any]] = []
        self._a_tasks: list[ActionTask] = []

        self.ctx = Context(-1, time.time())

    @tp.overload
    def add(
        self, interval: float, name: str, action: ActionFunction
    ) -> TaskHandle[tp.Any | None]: ...

    @tp.overload
    def add[T, S](
        self,
        name: str,
        source: SourceFunction[T],
        pipeline: PipelineFunction[T, S],
        sink: SinkFunction[S],
        *,
        policy: TaskPolicy=DEFAULT_TASK_POLICY,
    ) -> TaskHandle[T]: ...

    @tp.overload
    def add[T, S](
        self,
        interval: float,
        name: str,
        source: SourceFunction[T],
        pipeline: PipelineFunction[T, S],
        sink: SinkFunction[S],
        *,
        policy: TaskPolicy=DEFAULT_TASK_POLICY,
    ) -> TaskHandle[T]: ...

    def add(self, ni: float | str, *args, **kwargs) -> TaskHandle[tp.Any]:  # type: ignore[misc, no-untyped-def]
        """
        Add an asynchronous task to the scheduler.

        Three forms of the method are supported::

        `add(interval, name, action)`
            Add task to execute an asynchronous action.

        `add(interval, name, source, pipeline, sink)`
            Add data task to receive and process data at regular time
            intervals.

        `add(name, source, pipeline, sink)`
            Add data, event-driven task to receive and process data as it
            arrives.


        A data source function or an action can be

        - a coroutine
        - an asynchronous generator
        - an awaitable
        - a regular Python functions

        Data from `source` is buffered in a queue. Pipeline and sink are
        executed whenever there is data in the buffer. Data is lost when
        the queue is full.

        Pipeline and sink receive object of class :py:class:`n23.TaskResult`
        as its first parameter. Use `functools.partial` to set default
        parameters if the functions have more parameters.

        If `interval` is specified, then asynchronous task is run at
        regular intervals with a timer, thus obtaining equally spaced time
        series. Otherwise, task receives data as it arrives and unevenly
        spaced time series is generated.

        :param interval: Run task at regular time intervals (not defined
            for an event-driven task).
        :param name: Name (identifier) of the task. Unique for each task.
        :param source: Coroutine or function to receive source data
            (alternatively an action, see below).
        :param action: Coroutine or function to execute as an action.
        :param pipeline: Function to process source data (not defined for
            an action).
        :param sink: Coroutine to store or forward data (not defined for
            an action).
        :param policy: Task policy.

        .. seealso::

           - :py:class:`n23.TaskResult`
           - :py:class:`n23.TaskHandle`
        """
        num_args = len(args)
        logger.debug('add task: name or interval={}, num args={}'.format(
            ni, num_args
        ))

        match (ni, num_args):
            case (str(), int()):
                handle = self._add_e_task(ni, *args, **kwargs)
            case (float() | int(), 2):
                handle = self._add_action(ni, *args, **kwargs)
            case (float() | int(), int()):
                handle = self._add_r_task(ni, *args, **kwargs)
            case _:
                raise NotImplementedError('Method not implemented')
        return handle

    def _add_r_task[T, S](  # noqa: PLR0913
            self,
            interval: float,
            name: str,
            source: SourceFunction[T],
            pipeline: PipelineFunction[T, S],
            sink: SinkFunction[S],
            *,
            policy: TaskPolicy=DEFAULT_TASK_POLICY,
    ) -> TaskHandle[T]:
        """
        Submit task to execute at regular time intervals.
        """
        handle = TaskHandle[T](name)
        queue = NQueue[TaskResult[T]](max_size=policy.buffer_limit)
        task = RegularTask[T, S](
            handle, interval, source, pipeline, sink, queue
        )
        self._r_tasks.append(task)
        return handle

    def _add_e_task[T, S](
            self,
            name: str,
            source: SourceFunction[T],
            pipeline: PipelineFunction[T, S],
            sink: SinkFunction[S],
            *,
            policy: TaskPolicy=DEFAULT_TASK_POLICY,
    ) -> TaskHandle[T]:
        """
        Submit event-driven task to receive data as it arrives.
        """
        handle = TaskHandle[T](name)
        queue = NQueue[TaskResult[T]](max_size=policy.buffer_limit)
        task = EventDrivenTask[T, S](handle, source, pipeline, sink, queue)
        self._e_tasks.append(task)
        return handle

    def _add_action(
            self,
            interval: float,
            name: str,
            action: ActionFunction,
    ) -> TaskHandle[tp.Any]:
        """
        Submit action task to execute it at regular time intervals.

        :param interval: Run task at regular time intervals if specified.
        :param name: Name (identifier) of the task.
        :param action: Coroutine or function to execute as an action task.
        """
        handle = TaskHandle[tp.Any](name)
        task = ActionTask(handle, interval, action)
        self._a_tasks.append(task)
        return handle

    def __await__(self) -> Generator[tp.Any]:
        """
        Start scheduler and run tasks of the scheduler.
        """
        # always run the loop with regular and action tasks to generate
        # scheduler tick (even when there are no tasks)
        reg_task = asyncio.ensure_future(self._run_tasks())
        tasks: list[Awaitable[None]] = [reg_task]

        # add event-driven tasks
        tasks.extend(
            run_task_ed(t.handle, create_task_function(t.source), t.queue)
            for t in self._e_tasks
        )

        # add task data processors using pipelines and sinks
        items: Iterable[DataTask] = concat(self._r_tasks, self._e_tasks)
        tasks.extend(
            task_process_data(t.queue, t.pipeline, t.sink) for t in items
        )
        yield from asyncio.gather(*tasks).__await__()

    async def _run_tasks(self) -> None:
        """
        Run N23 scheduler tasks executed at regular time intervals.
        """
        task_cf: tp.Callable[[RegularTask[tp.Any, tp.Any]], TaskCoroutineFunction[None]] = \
            lambda t: partial(
                run_task_regular,
                t.handle, create_task_function(t.source), t.queue
            )
        tasks = [(t.interval, t.handle, task_cf(t)) for t in self._r_tasks]
        tasks.extend(
            (t.interval, t.handle, create_task_function(t.action))
            for t in self._a_tasks
        )

        intervals = set(i for i, *_ in tasks)
        if len(intervals) > 1:
            # TODO: add support for multiplies of the main interval
            raise ValueError('Single interval is supported only')
        elif not intervals:
            interval = 1.0
        else:
            interval = intervals.pop()
        assert not intervals

        timer = Timer(interval)
        timeout = interval * 0.75 if self.timeout is None else self.timeout

        ctx = self.ctx
        wait_tasks = partial(
            asyncio.wait,
            timeout=timeout,
            return_when=asyncio.FIRST_EXCEPTION
        )
        create_task = asyncio.create_task
        exclude = set[str]()
        pending = set[asyncio.Task[tp.Any]]()
        try:
            timer.start()
            while True:
                await timer
                ctx.time = time.time()
                ctx.tick = timer._tick  # noqa: SLF001

                current_tasks = [
                    create_task(cf(), name=h.name) for _, h, cf in tasks
                    if h.name not in exclude
                ]
                current_tasks.extend(pending)
                if current_tasks:
                    done, pending = await wait_tasks(current_tasks)

                    # crash software on error
                    errors = [e for t in done if (e := t.exception())]
                    if errors:
                        raise ExceptionGroup('Task execution failed', errors)

                    exclude = {t.get_name() for t in pending}
                    if exclude:
                        logger.info('pending tasks: n={}, tasks(5)={}'.format(
                            len(exclude), ', '.join(islice(exclude, 5))
                        ))
        finally:
            timer.close()

#
# functions to create task coroutines, execute tasks, and process task data
# for N23 scheduler
#

def create_task_function[T](
        func: SourceFunction[T] | ActionFunction
) -> TaskCoroutineFunction[tp.Any]:
    """
    Create asynchronous coroutine function to execute a task of N23
    scheduler.

    A data source function or an action can be

    - a coroutine
    - an asynchronous generator
    - an awaitable
    - a regular Python functions

    :param func: Data source function or an action.
    """

    task_cf: TaskCoroutineFunction[tp.Any]
    if inspect.isawaitable(func):
        task_cf = lambda: func  # type: ignore[assignment, return-value]
        logger.info('{} is awaitable'.format(func))
    elif inspect.iscoroutinefunction(func):
        task_cf = func
        logger.info('{} is coroutine function'.format(func))
    elif inspect.isasyncgenfunction(func):
        task_cf = func().__anext__
        logger.info('{} is asynchronous generator'.format(func))
    else:
        task_cf = partial(asyncio.to_thread, func)
        logger.info('{} is function, run in a thread' .format(func))

    return task_cf

async def run_task_regular[T](
        handle: TaskHandle[T],
        task_cf: TaskCoroutineFunction[T],
        queue: NQueue[TaskResult[T]],
) -> None:
    """
    Execute task and put result of the task into the buffer queue.

    :param handle: N23 scheduler task handle.
    :param task_cf: Task coroutine function.
    :param queue: N23 scheduler task buffer queue.
    """
    value = await task_cf()
    result = TaskResult(time.time(), handle.name, value)

    if __debug__ and queue.full():
        logger.debug('queue to drop task results: task={}'.format(
            handle.name
        ))
    queue.put(result)

async def run_task_ed[T](
        handle: TaskHandle[T],
        task_cf: TaskCoroutineFunction[T],
        queue: NQueue[TaskResult[T]],
) -> None:
    """
    Execute N23 scheduler event-driven task and process its data as it
    arrives.

    :param handle: N23 scheduler task handle.
    :param task_cf: Task coroutine function.
    :param queue: N23 scheduler task buffer queue.
    """
    while True:
        await run_task_regular(handle, task_cf, queue)

async def task_process_data[T, S](
        queue: NQueue[TaskResult[T]],
        pipeline: PipelineFunction[T, S],
        sink: SinkFunction[S],
) -> None:
    while True:
        value = await queue.get()
        if (result := pipeline(value)) is not None:
            await sink(result)

__all__ = ['Scheduler', 'TaskHandle', 'TaskResult']

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2024 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import hy  # noqa: F401

from .config import read_app_config, app_config_value
from .core import run
from .op import n23_add, n23_scheduler, n23_sink_from_uri, n23_process  # type: ignore[attr-defined]
from .runner import run_app
from .types import ApplicationConfig, ApplicationConfigError

__all__ = [
    'ApplicationConfig',
    'ApplicationConfigError',
    'app_config_value',
    'n23_add',
    'n23_process',
    'n23_scheduler',
    'n23_sink_from_uri',
    'read_app_config',
    'run',
    'run_app',
]

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2024 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os.path
import typing as tp
from collections.abc import Iterator
from datetime import datetime, timezone
from contextlib import contextmanager
from importlib import resources

logger = logging.getLogger(__name__)

T = tp.TypeVar('T')

def to_datetime(ts: float) -> datetime:
    """
    Convert Unix epoch to datetime object in UTC timezone.

    :param ts: Unix epoch.
    """
    return datetime.fromtimestamp(ts, tz=timezone.utc)

@contextmanager
def read_file(fn: str, ext: str) -> Iterator[str]:
    """
    Read content of a file or a module.

    :param fn: File or resource path.
    """
    if not os.path.exists(fn):
        pkg, name = fn.rsplit('.', 1)
        fn = resources.files(pkg).joinpath(f'{name}.{ext}')  # type: ignore

    logger.info(f'reading file: {fn}')
    with open(fn) as f:
        yield f.read(65 * 1024)  # TODO: make it configurable

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2025 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Implementation of default data storage for N23 framework.

The storage is backed by PostgreSQL RDBMS. It is tested with TimescaleDB
and PostGIS extensions.

The storage is implemented with :py:class:`Database` class.

A database table, which stores application data, needs to be declared with
:py:meth:`Database.add_entity` method. It also allows to specify table
columns.

For example::

    db.add_entity('temperature`, ('location', 'value'))

and for task results::

    TaskResult(time=10001, name='temperature/room1', value=20.1)
    TaskResult(time=10002, name='temperature/room1', value=20.2)
    TaskResult(time=10003, name='temperature/room1', value=20.3)

the data is stored in table `temperature` as::

    +-------+----------+-------+
    | time  | location | value |
    +-------+----------+-------+
    | 10001 | room1    |  20.1 |
    | 10002 | room1    |  20.2 |
    | 10003 | room1    |  20.3 |
    +-------+----------+-------+

"""

import abc
import asyncio
import asyncpg  # type: ignore
import logging
import types
import typing as tp
from collections import deque
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from functools import singledispatch, partial
from urllib.parse import urlparse, urlunparse

from ..api import DEFAULT_MAX_LEN, Storage, Fields
from ...fn import identity
from ...scheduler import TaskResult
from ...util import to_datetime

logger = logging.getLogger(__name__)

T_co = tp.TypeVar('T_co', covariant=True)
TypeEncoder: tp.TypeAlias = tp.Callable[..., tp.Any]
TypeEncoders: tp.TypeAlias = dict[str, TypeEncoder]
TableRow: tp.TypeAlias = tuple[tp.Any, ...]

# see https://github.com/MagicStack/asyncpg/issues/513
CONNECTION_ERRORS = (
    OSError,
    asyncpg.CannotConnectNowError,
    asyncpg.ConnectionDoesNotExistError,
)

CONNECTION_ERROR_BACK_TO_POOL = 'connection has been released back to the pool'
CONNECTION_ERROR_CLOSED = 'the underlying connection is closed'

SLEEP_ON_CONNECTION_ERROR = 5

class Database(Storage):
    """
    N23 framework storage implementation using PostgreSQL RDBMS.
    """
    def __init__(self, db_uri: str) -> None:
        """
        Initialize storage backend.

        :param db_uri: Database URI specification.
        """
        self.db_uri = db_uri

        # connection pool size controls how many parallel connections are
        # performed
        self.pool_size = 1

        self._type_codecs: TypeEncoders = {}
        self._writers: dict[str, TableWriterBase] = {}

        self._pool = asyncpg.create_pool(
            self.db_uri,
            min_size=1,
            max_size=self.pool_size,
            # TODO: use context vars instead of partial application
            init=partial(set_connection_codecs, self._type_codecs)
        )

    def add_entity(
            self,
            name: str,
            fields: Fields,
            *,
            batch: bool=False,
            max_len: int=DEFAULT_MAX_LEN,
    ) -> None:
        """
        Add and configure storage entity - PostgreSQL database table.
        """
        if name in self._writers:
            raise ValueError('Entity "{}" already defined' .format(name))
        if len(fields) < 1:
            raise ValueError('At least one column expected')

        cols = ('time', *fields)
        writer: TableWriterBase
        if batch:
            writer = TableBatchWriter(self._pool, name, cols, max_len)
        else:
            writer = TableWriter(self._pool, name, cols)

        self._writers[name] = writer

    async def write(self, result: TaskResult[T_co]) -> None:
        """
        Write N23 framework task result into database table.

        :param result: Task result generated by N23 scheduler.
        """
        table, *name_values = result.name.split('/')
        ts = to_datetime(result.time)
        writer = self._writers[table]

        # is single or multiple values expected in task result?
        # compare number of table declared columns with number of values to
        # be inserted into the table
        # - time column
        # - values extracted from the task result name
        # - at least one value from the task result
        is_single = len(writer._columns) == len(name_values) + 2  # noqa: SLF001
        values = (result.value,) if is_single else result.value

        row = (ts, *name_values, *values)  # type: ignore
        await writer.write(row)

    def set_type_codec(self, type_name: str, encoder: TypeEncoder) -> None:
        """
        Assign encoder to a database type.

        :param type_name: Database type name.
        :param encoder: Function converting Python value to format understood
            by database.
        """
        self._type_codecs[type_name] = encoder

    async def flush(self) -> None:
        tasks = (
            w.flush() for w in self._writers.values()
            if hasattr(w, 'flush')
        )
        task = asyncio.gather(*tasks)
        await catch_conn_error(task)

    async def __aenter__(self) -> None:
        await self._pool

    async def __aexit__(
        self,
        ex_type: type[BaseException] | None,
        ex: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        try:
            async with asyncio.timeout(30):
                await self.flush()
                await self._pool.close()
        except asyncio.TimeoutError:
            logger.warning('database flush failed due to timeout: %s', ex)

class TableWriterBase:
    """
    Basic class for PostgreSQL database table writers.
    """
    def __init__(
            self,
            pool: asyncpg.Pool,
            table: str,
            columns: Fields,
    ):
        self._pool = pool
        self._table = table
        self._columns = columns

    @abc.abstractmethod
    async def write(self, row: TableRow) -> None:
        ...

class TableWriter(TableWriterBase):
    """
    PostgreSQL database table writer.
    """
    async def write(self, row: TableRow) -> None:
        """
        Writer row to a PostgreSQL database table.

        :param row: Data to write.
        """
        while True:
            task = self._write(row)
            saved = await catch_conn_error(task)
            if saved:
                break
            else:
                await asyncio.sleep(SLEEP_ON_CONNECTION_ERROR)

    async def _write(self, row: TableRow) -> None:
        async with connect(self._pool) as conn:  # type: ignore
            await conn.copy_records_to_table(
                self._table,
                records=[row],
                columns=self._columns,
            )

        if __debug__:
            logger.debug('stored 1 record in table {}'.format(self._table))

class TableBatchWriter(TableWriterBase):
    """
    PostgreSQL database table batch writer.

    Data is added to an internal queue on write. Data is flushed to
    database table using :py:meth:`TableBatchWriter.flush` method.
    """
    def __init__(
            self,
            pool: asyncpg.Pool,
            table: str,
            columns: Fields,
            max_len: int,
    ):
        super().__init__(pool, table, columns)
        self._queue = deque[TableRow]([])
        self._lock = asyncio.Lock()
        self._max_len = max_len

    async def write(self, row: TableRow) -> None:
        """
        Enqueue row of data to be written into a PostgreSQL database table.

        :param row: Data to write.
        """
        # use lock to avoid error, when asyncpg iterates over the table
        # rows while saving into database
        async with self._lock:
            self._queue.append(row)

        if len(self._queue) == self._max_len:
            await self.flush()

    async def flush(self) -> None:
        """
        Flush enqueued rows of data into a PostgreSQL database table.

        Do nothing if internal queue is empty.
        """
        if not self._queue:
            return

        async with self._lock, connect(self._pool) as conn:  # type: ignore
            await conn.copy_records_to_table(
                self._table,
                records=self._queue,
                columns=self._columns,
            )
            count_rows = len(self._queue)
            self._queue.clear()

        if __debug__:
            logger.debug(
                'stored {} record(s) in table {}'.format(count_rows, self._table)
            )

async def catch_conn_error(coro: tp.Awaitable[tp.Any]) -> bool:
    """
    Protect calling coroutine from connection error of coroutine being
    executed.

    Return true if coroutine executed without connection error.

    :param coro: Coroutine to execute.
    """
    try:
        await coro
    except CONNECTION_ERRORS as ex:
        logger.warning(
            'connection failure due to: {}'.format(ex)
        )
        return False
    except asyncpg.InterfaceError as ex:
        if not is_disconnected(ex):
            raise
        logger.warning(
            'connection failure due to: {}'.format(ex)
        )
        return False
    else:
        return True

@asynccontextmanager
async def create_pool(
        db_uri: str,
        type_codecs: TypeEncoders={},
        max_size: int=1,
    ) -> AsyncIterator[asyncpg.Pool]:
    """
    Create connection pool for a database.

    :param db_uri: Database URI.
    :param type_codecs: Mapping between database types and their encoders.
    :param max_size: Maximum number of connections of connection pool.
    """
    async with asyncpg.create_pool(
                db_uri,
                min_size=1,
                max_size=max_size,
                init=partial(set_connection_codecs, type_codecs),
            ) as pool:

        yield pool

@singledispatch  # type: ignore
@asynccontextmanager
async def connect(
        db: str | asyncpg.Pool,
        type_codecs: TypeEncoders={},
        max_size: int=1,
        enable_tx: bool=True
    ) -> AsyncIterator[asyncpg.Connection]:
    """
    Context manager to connect to a database.

    First parameter can be

    - database URI
    - or database connection pool

    In first instance, new connection pool is created to connect to
    a database. In second instance, the connection pool is reused to
    connect to a database.

    :param db: Database URI or database connection pool.
    :param max_size: Maximum number of connections of connection pool.
    :param enable_tx: Create database transaction (true by default).
    """
    raise NotImplementedError(
        'Not implemented for database parameter type: {}'
        .format(type(db))
    )

@connect.register
@asynccontextmanager
async def _connect_db_uri(
        db_uri: str,
        type_codecs: TypeEncoders={},
        max_size: int=1,
        enable_tx: bool=True
    ) -> AsyncIterator[asyncpg.Connection]:
    """
    Create connection pool and use it to connect to a database.
    """
    async with create_pool(db_uri, type_codecs, max_size=max_size) as pool:
        async with connect(pool, enable_tx=enable_tx) as conn:  # type: ignore
            yield conn

@connect.register
@asynccontextmanager
async def _connect_db_pool(
        pool: asyncpg.Pool,
        enable_tx: bool=True
    ) -> AsyncIterator[asyncpg.Connection]:
    """
    Reuse existing connection pool to connect to a database.
    """
    async with pool.acquire() as conn:
        if enable_tx:
            async with conn.transaction():
                yield conn
        else:
            yield conn

async def set_connection_codecs(
        type_codecs: TypeEncoders,
        conn: asyncpg.Connection,
    ) -> None:
    set_codec = partial(
        conn.set_type_codec, decoder=identity, format='binary'
    )
    for t, f in type_codecs.items():
        if __debug__:
            logger.debug('set codec for type {}: {}'.format(t, f))
        await set_codec(t, encoder=f)

async def execute(db_uri: str, cmd: str, enable_tx: bool=True) -> None:
    """
    Connect to database and execute single database statement.
    """
    logger.info('execute: {}'.format(cmd))
    async with connect(db_uri, enable_tx=enable_tx) as conn:  # type: ignore
        await conn.execute(cmd)

def is_disconnected(ex: asyncpg.InterfaceError) -> bool:
    """
    Check if the PostgreSQL interface error is a disconnection error.
    """
    message = str(ex)
    disconnected = CONNECTION_ERROR_BACK_TO_POOL in message \
        or CONNECTION_ERROR_CLOSED in message
    logger.debug(
        'interface error (disconnected={}): {}'
        .format(disconnected, message)
    )
    return disconnected

def to_db_uri(db: str, alt_db: tp.Optional[str]=None) -> str:
    """
    Convert database name or partial database connection string to database
    connection string for local database server.

    If the `db` parameter is database connection string already, then do
    nothing.

    For example::

        some-db -> postgresql://localhost/some-db
        localhost/some-db -> postgresql://localhost/some-db
        postgresql://localhost/some-db -> postgresql://localhost/some-db
        user:pass@localhost/some-db -> postgresql://user:pass@localhost/some-db

    Use alternative database name to get connection for template database,
    i.e. `template1`.

    @param db: Database name or database connection string.
    @param alt_db: Alternative database name.
    """
    db = db if db.startswith('postgresql://') or db.startswith('//') else '//' + db
    db_uri = urlparse(db, scheme='postgresql')

    # if no hostname in `db`, then netloc points to dbname, therefore make
    # a swap and use localhost for hostname
    host, dbname = (db_uri.netloc, db_uri.path) \
        if db_uri.netloc and db_uri.path \
        else ('localhost', db_uri.netloc)
    dbname = dbname if alt_db is None else alt_db

    db_uri = db_uri._replace(path=dbname, netloc=host)
    logger.info('parsed db uri: {}'.format(db_uri))
    return urlunparse(db_uri)

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2025 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
API definitions for N23 framework data storage.

The API strives to be SQL/RDBMS independent.

Requirements for a storage backend

- if storage is an external service, like database server, the storage
  write method should block or enqueue data
- even if database server is down forever, a storage implementation must
  not cause an application to run out of memory
"""

import abc
from collections.abc import Awaitable
import types
import typing as tp
from contextlib import AbstractAsyncContextManager
from urllib.parse import urlparse

from ..error import ApplicationError
from ..scheduler import TaskResult

DEFAULT_MAX_LEN = 300

T_co = tp.TypeVar('T_co', covariant=True)

# a collection of fields of a data record, i.e. table column names
Fields: tp.TypeAlias = tp.Sequence[str]

class Storage(AbstractAsyncContextManager[Awaitable[tp.Any] | None], tp.Protocol):
    """
    N23 framework storage protocol.
    """
    @abc.abstractmethod
    def add_entity(
            self,
            name: str,
            fields: Fields,
            *,
            batch: bool=False,
            max_len: int=DEFAULT_MAX_LEN,
    ) -> None:
        """
        Add and configure storage entity.

        An example of storage entity is a database table.
        """

    @abc.abstractmethod
    async def write(self, result: TaskResult[T_co]) -> None:
        """
        Write result of N23 task into storage.

        Storage needs to be configured via :py:meth:`Database.add_entity`
        method.

        :param result: Task result generated by N23 scheduler.
        """

    @abc.abstractmethod
    async def flush(self) -> None:
        """
        Flush data buffered for write.
        """

    async def __aenter__(self) -> None:
        return None

    async def __aexit__(
        self,
        ex_type: type[BaseException] | None,
        ex: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        return None

class StorageAbyss(Storage):
    """
    Storage entity, which does nothing.
    """
    def add_entity(
            self,
            name: str,
            fields: Fields,
            *,
            batch: bool=False,
            max_len: int=DEFAULT_MAX_LEN,
    ) -> None:
        pass

    async def write(self, result: TaskResult[T_co]) -> None:
        pass

    async def flush(self) -> None:
        pass


def storage_from_uri(uri: str) -> Storage:
    """
    Create storage object or connection for N23 framework.
    """
    v = urlparse(uri)
    f = STORAGE.get(v.scheme)
    if f is None:
        raise ApplicationError('Unknown storage: {}'.format(v.scheme))

    return f(uri)

def _pg_from_uri(uri: str) -> Storage:
    """
    Create PostgreSQL or TimescaleDB storage object using URI.
    """
    # pylint: disable=import-outside-toplevel
    import n23.storage.pg
    return n23.storage.pg.Database(uri)

STORAGE = {
    'postgresql': _pg_from_uri,
}

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2023 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ruff: noqa: SLF001

"""
The n23 scheduler unit tests.
"""

import asyncio
import functools
import logging
import time
import typing as tp
from collections import Counter
from collections.abc import AsyncIterator, Awaitable

from n23.fn import identity
from n23.scheduler import create_task_function, run_task_ed, Scheduler, \
    TaskHandle, TaskResult, logger as s_logger
from n23.queue import NQueue

import pytest
from unittest import mock

logger = logging.getLogger()

@pytest.mark.asyncio
async def test_task_read_awaitable() -> None:
    """
    Test reading data with an awaitable.
    """
    class Reader(Awaitable[str]):
        def __await__(self) -> tp.Generator[None, None, str]:
            yield from ()
            return 'value'

    cf = create_task_function(Reader())
    assert await cf() == 'value'

@pytest.mark.asyncio
async def test_task_read_func() -> None:
    """
    Test reading data with a function.
    """
    cf = create_task_function(lambda: 'value')
    assert await cf() == 'value'

@pytest.mark.asyncio
async def test_task_read_coroutine() -> None:
    """
    Test reading data with a coroutine.
    """
    async def reader() -> str:
        return 'value'

    cf = create_task_function(reader)
    assert await cf() == 'value'

@pytest.mark.asyncio
async def test_task_read_coroutine_partial() -> None:
    """
    Test reading data with coroutine enclosed with partial.
    """
    async def reader(v: str) -> str:
        return 'value ' + v

    cf = create_task_function(functools.partial(reader, 'test'))
    assert await cf() == 'value test'

@pytest.mark.asyncio
async def test_task_read_async_generator() -> None:
    """
    Test reading data with a asynchronous generator.
    """
    async def reader() -> AsyncIterator[str]:
        while True:
            yield 'value'

    cf = create_task_function(reader)
    assert await cf() == 'value'

@pytest.mark.timeout(1)
@pytest.mark.asyncio
async def test_task_ed_read() -> None:
    """
    Test reading data with event-driven task.
    """
    async def reader() -> AsyncIterator[str]:
        for i in range(3):
            await asyncio.sleep(0.001)
            yield 'value'

    th = TaskHandle[str]('pressure')
    queue = NQueue[TaskResult[str]](5)
    task = run_task_ed(th, create_task_function(reader), queue)

    try:
        await task
    except StopAsyncIteration as ex:
        logger.info('ignoring: {}'.format(ex))

    result = queue._data
    assert [v.name for v in result] == ['pressure'] * 3
    assert [v.value for v in result] == ['value'] * 3

def test_scheduler_adding_regular_task() -> None:
    """
    Test adding regular task with pipeline and sink to scheduler.
    """
    async def source() -> int: return 1
    async def sink(v: TaskResult[int]) -> None: pass

    scheduler = Scheduler()
    handle = scheduler.add(1, 'pressure', source, identity, sink)
    assert not scheduler._e_tasks
    assert not scheduler._a_tasks

    task = scheduler._r_tasks[0]
    assert task.interval == 1
    assert task.handle == handle
    assert task.source == source
    assert task.pipeline == identity
    assert task.sink == sink

def test_scheduler_adding_ed_task() -> None:
    """
    Test adding event-driven task with pipeline and sink to scheduler.
    """
    async def source() -> int: return 1
    async def sink(v: TaskResult[int]) -> None: pass

    scheduler = Scheduler()
    handle = scheduler.add('pressure', source, identity, sink)
    assert not scheduler._r_tasks
    assert not scheduler._a_tasks

    task = scheduler._e_tasks[0]
    assert task.handle == handle
    assert task.source == source
    assert task.pipeline == identity
    assert task.sink == sink

def test_scheduler_adding_action_task() -> None:
    """
    Test adding action task with pipeline and sink to scheduler.
    """
    async def action() -> None: pass

    scheduler = Scheduler()
    handle = scheduler.add(1, 'pressure', action)
    assert not scheduler._r_tasks
    assert not scheduler._e_tasks

    task = scheduler._a_tasks[0]
    assert task.interval == 1
    assert task.action == action
    assert task.handle == handle

@pytest.mark.timeout(5)
@pytest.mark.asyncio
async def test_scheduler_reader_regular() -> None:
    """
    Test scheduler reading data with a coroutine.
    """
    async def reader() -> int:
        return 2

    sink = mock.AsyncMock(side_effect=[None, ValueError('stop')])

    start = time.time()
    scheduler = Scheduler()
    scheduler.add(0.1, 'pressure', reader, identity, sink)
    with pytest.raises(ValueError, match='^stop$'):
        await scheduler

    assert scheduler.ctx.tick == 1
    assert scheduler.ctx.time > start

    # data is consumed by the sink
    assert len(scheduler._r_tasks[0].queue._data) == 0

@pytest.mark.asyncio
async def test_scheduler_reader_error() -> None:
    """
    Test scheduler not swallowing error thrown by data reader.
    """
    async def reader() -> tp.NoReturn:
        raise ValueError('test test test')

    sink = mock.AsyncMock()

    scheduler = Scheduler()
    scheduler.add(0.1, 'pressure', reader, identity, sink)
    with pytest.raises(ExceptionGroup, match='^Task execution failed') as ctx:
        await scheduler
    assert ctx.group_contains(ValueError, match='^test test test$')

@pytest.mark.timeout(1)
@pytest.mark.asyncio
async def test_scheduler_reader_timeout() -> None:
    """
    Test scheduler logging information on data read timeout.
    """
    with mock.patch.object(asyncio, 'wait') as mock_wait:
        task = mock.MagicMock()
        task.get_name.return_value = 'pressure'

        stopper = mock.MagicMock()
        stopper.exception.return_value = ValueError('stop')
        mock_wait.side_effect = [
            [{}, {task}],
            [{stopper}, {}],
        ]

        scheduler = Scheduler()
        scheduler.add(
            0.1, 'pressure', mock.MagicMock(),  identity, mock.MagicMock()
        )
        p_logger = mock.patch.object(s_logger, 'info')
        with p_logger as f, \
                pytest.raises(ExceptionGroup, match='^Task execution failed') as ctx:
            await scheduler._run_tasks()

        assert ctx.group_contains(ValueError, match='^stop$')

        # the timeout was logged and task got cancelled
        c1 = f.call_args_list[1]
        assert c1 == mock.call('pending tasks: n=1, tasks(5)=pressure')

@pytest.mark.timeout(1)
@pytest.mark.asyncio
async def test_scheduler_pipeline_null() -> None:
    """
    Test scheduler skipping sink on pipeline returing null.
    """
    async def reader() -> AsyncIterator[int]:
        yield 1
        yield 1
        raise ValueError('stop')

    counter = Counter[int]()
    def pl(v: TaskResult[int]) -> TaskHandle[int] | None:
        counter[v.value] += 1
        return None

    sink = mock.AsyncMock()

    scheduler = Scheduler()
    scheduler.add(0.1, 'pressure', reader, pl, sink)
    assert len(scheduler._r_tasks) == 1

    with pytest.raises(ExceptionGroup, match='^Task execution failed') as ctx:
        await scheduler
    assert ctx.group_contains(ValueError, match='^stop$')

    assert counter[1] == 2  # pipeline called twice
    assert sink.call_count == 0  # but sink is not called at all

@pytest.mark.timeout(1)
@pytest.mark.asyncio
async def test_scheduler_sink_error() -> None:
    """
    Test scheduler not swallowing error thrown by sink.
    """
    async def reader() -> int:
        return 1

    sink = mock.AsyncMock()
    sink.side_effect = [ValueError('test test test')]

    scheduler = Scheduler()
    scheduler.add(0.1, 'pressure', reader, identity, sink)
    assert len(scheduler._r_tasks) == 1

    with pytest.raises(ValueError, match='^test test test$'):
        await scheduler

    sink.assert_called_once()

@pytest.mark.timeout(3)
@pytest.mark.asyncio
async def test_scheduler_cancel() -> None:
    """
    Test cancelling scheduler.
    """
    async def reader() -> int:
        return 1

    async def reader_event() -> int:
        await asyncio.sleep(0.25)
        return 1

    async def cancel(task: asyncio.Future[None]) -> None:
        await asyncio.sleep(1)
        task.cancel('done')

    sink = mock.AsyncMock()
    sink_event = mock.AsyncMock()

    scheduler = Scheduler()
    scheduler.add(0.2, 'pressure', reader, identity, sink)
    scheduler.add('gps', reader_event, identity, sink_event)

    task = asyncio.ensure_future(scheduler)
    with pytest.raises(asyncio.CancelledError, match='^done$'):
        await asyncio.gather(cancel(task), task)

    # sleep a while and check if all tasks are cancelled
    await asyncio.sleep(1)

    # expected 5, but give it a bit of flexibility
    assert 4 <= sink.call_count <= 6

    # expected 4, but give it a bit of flexibility
    assert 3 <= sink_event.call_count <= 6

    assert len(asyncio.all_tasks()) == 1

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2024 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import dataclasses as dtc
import typing as tp
from collections.abc import Awaitable
from contextlib import AbstractAsyncContextManager

from ..scheduler import Scheduler

Process: tp.TypeAlias = Awaitable[tp.Any]

@dtc.dataclass(frozen=True)
class ApplicationRunContext:
    """
    N23 application run context.

    NOTE: Process manager should return null or an awaitable.

    :var setup: Function to setup/start application to be run via N23
        framework.
    :var scheduler: N23 scheduler.
    :var process_managers: Context managers, which create processes.
    :var processes: List of processes.
    """
    setup: tp.Callable[[], Awaitable[tp.Any]]
    scheduler: Scheduler
    process_managers: list[AbstractAsyncContextManager[Process]]
    processes: list[Process]

@dtc.dataclass(frozen=True)
class ApplicationConfigItem:
    name: str
    value: object

@dtc.dataclass(frozen=True)
class ApplicationConfigSection:
    name: str
    items: list[ApplicationConfigItem]

@dtc.dataclass(frozen=True)
class ApplicationConfig:
    section: list[ApplicationConfigSection]

class ApplicationConfigError(Exception):
    """
    N23 application configuration error.
    """

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2024 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Collection of functions to support functional programming in Python.
"""

import typing as tp
from collections.abc import Iterable
from inspect import signature
from itertools import chain, islice
from functools import partial

T = tp.TypeVar('T')
S = tp.TypeVar('S')

concat = chain
flatten = chain.from_iterable

def identity(v: T) -> T:
    """
    Identity function.
    """
    return v

def head(it: Iterable[T]) -> T:
    """
    Get head (first) item of an iterable.
    """
    return next(islice(it, 0, 1))

def tail(it: Iterable[T]) -> Iterable[T]:
    """
    Get tail (all items, but the first) of an iterable.
    """
    return islice(it, 1, None)

def compose(*funcs):  # type: ignore
    """
    Compose functions.
    """
    first = head(funcs)
    rest = tail(funcs)

    first_has_param = bool(signature(first).parameters)

    def wrapper_0():  # type: ignore
        result = first()
        for f in rest:
            result = f(result)
        return result

    def wrapper_1(v):  # type: ignore
        result = first(v)
        for f in rest:
            result = f(result)
        return result

    return wrapper_1 if first_has_param else wrapper_0

__all__ = [
    'compose', 'concat', 'flatten', 'head', 'identity', 'partial', 'tail',
]

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2024 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import dataclasses as dtc
import hy
import logging
import typing as tp
from functools import partial
from inspect import get_annotations
from hy.model_patterns import Expression, Symbol
from hy.models import Object

logger = logging.getLogger(__name__)

T = tp.TypeVar('T')
S = tp.TypeVar('S')

def parse_dsl(
        data: str, parser: Expression, cls: type[T], /, positional: bool=False
) -> T:
    form = list(hy.read_many(data))
    qform = [qquote(s) if is_toplevel(cls, s) else s for s in form]
    eform = [s for s in hy.eval(qform) if s is not None]
    if positional:
        return parse_struct_pos(cls, parser.parse(eform))
    else:
        args = parse_attrs_as_dict(cls, parser.parse(eform))
        return cls(**args)

def qquote(expr: Expression) -> Expression:
    return Expression([Symbol('quasiquote'), expr])

def is_toplevel(cls: type, expr: Expression) -> bool:
    name = str(expr[0])
    fields = {f.name for f in dtc.fields(cls)}
    return name in fields

def parse_attr_name(attr: Symbol | str) -> str:
    return hy.mangle(str(attr))  # type: ignore[no-any-return]

def get_attr_type(cls: type, attr: str) -> type:
    name = parse_attr_name(attr)
    return get_annotations(cls)[name]  # type: ignore[no-any-return]

def parse_attr_value(attr_type: type[T], attr_value: tp.Any) -> tp.Any:

    attr_sub_type = getattr(attr_type, '__args__', [None])[0]

    logger.debug('parsing value, type={}, sub-type={}, value={}'.format(
        attr_type, attr_sub_type, attr_value
    ))

    result: tp.Any
    match attr_value:
        case Expression():
            result = [
                parse_attr_value(attr_sub_type, item) for item in attr_value
            ]
        case Symbol():
            result = str(attr_value)
        case Object():
            result = hy.eval(attr_value)
        case [Symbol(), *_]:
            result = tuple(
                parse_attr_value(attr_sub_type, item) for item in attr_value
            )
        case _:
            result = attr_value

    return result

def parse_attrs_as_dict(cls: type[T], args: list[tp.Any]) -> dict[str, tp.Any]:
    items = (v for v in args if v is not None)
    to_type = partial(get_attr_type, cls)
    return {
        parse_attr_name(key): parse_attr_value(to_type(key), value)
        for key, value in items
    }

def parse_struct_kw(cls: type[T], args: list[tp.Any]) -> tuple[str, T]:
    kw = parse_attrs_as_dict(cls, args[1:])
    return (args[0], cls(**kw))

def parse_struct_kw_named(cls: type[T], args: list[tp.Any]) -> T:
    name = str(args[0])
    kw = parse_attrs_as_dict(cls, args[1:])
    return cls(name, **kw)  # type: ignore

def parse_struct_pos(cls: type[T], args: list[tp.Any]) -> T:
    fields = (f.name for f in dtc.fields(cls))  # type: ignore
    items = ((get_attr_type(cls, n), v) for n, v in zip(fields, args))
    values: list[tp.Any] = [parse_attr_value(t, v) for t, v in items]
    return cls(*values)

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2024 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Queue with limited number of items.

Adding new items to the queue removes the oldest one when the queue is
full.
"""

import asyncio
from collections import deque

class NQueue[T]:
    """
    Queue with limited number of items.

    Adding an item to a full queue drops oldest queue item.
    """
    def __init__(self, max_size: int) -> None:
        """
        Create queue with its size specified.

        :param max_size: Size of the queue.
        """
        self.max_size = max_size
        self._data = deque[T]([], maxlen=max_size)
        self._task: asyncio.Future[T] | None = None

    def put(self, item: T) -> None:
        """
        Add an item to the queue.

        :param item: Item to add to the queue.
        """
        self._data.append(item)
        task = self._task
        if task is not None and not task.done() and self._data:
            task.set_result(self._data.popleft())

    async def get(self) -> T:
        """
        Get the oldest item from the queue.

        Wait if queue is empty.
        """
        if self._task is not None:
            raise TypeError('A consumer already registered for the queue')

        if self._data:
            result = self._data.popleft()
        else:
            loop = asyncio.get_running_loop()
            self._task = loop.create_future()
            result = await self._task
            self._task = None

        return result

    def full(self) -> bool:
        """
        Return true if queue is full.
        """
        return len(self._data) == self._data.maxlen

    def __len__(self) -> int:
        """
        Return current size of the queue.
        """
        return len(self._data)

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2023 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import dataclasses as dtc

@dtc.dataclass(frozen=True)
class Storage:
    version: str
    role: str
    extensions: tuple[str, ...]=tuple()

@dtc.dataclass(frozen=True)
class Column:
    name: str
    type: str
    nullable: bool

@dtc.dataclass(frozen=True)
class Entity:
    name: str
    partition_by: str | None=None
    columns: list[Column]=dtc.field(default_factory=list)
    chunk: str='1day'

@dtc.dataclass(frozen=True)
class Config:
    storage: Storage
    entities: list[Entity]
    sql: str=''

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2024 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from n23.util import to_datetime

import pytest

DATETIME_DATA = (
    (1654038062, '2022-05-31 23:01:02+00:00'),
    (1667260873, '2022-11-01 00:01:13+00:00'),
)

@pytest.mark.parametrize('ts, expected', DATETIME_DATA)
def test_datetime_conv(ts: float, expected: str) -> None:
    """
    Test Unix epoch to datetime conversion.
    """
    dt = to_datetime(ts)
    assert dt.isoformat(sep=' ') == expected

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2023 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import typing as tp

from ..dsl import parse_dsl
from ..util import read_file
from .config_schema import PARSER
from .types import ApplicationConfig, ApplicationConfigError

def read_app_config(fn: str) -> ApplicationConfig:
    """
    Read N23 application configuration file.

    :param fn: File or resource path.
    """
    with read_file(fn, 'cfg') as data:
        return parse_dsl(data, PARSER, ApplicationConfig, positional=True)

def app_config_value(
        config: ApplicationConfig, section: str, name: str, default: tp.Any=None
) -> tp.Any:
    """
    Get configuration value for N23 application.

    :param config: N23 application configuration.
    :param section: Section name of the configuration.
    :param name: Configuration item value.
    """
    no_value = object()
    items = (
        c.value for s in config.section for c in s.items
        if (s.name, c.name) == (section, name)
    )
    value = next(items, no_value)
    if default is None and value is no_value:
        raise ApplicationConfigError(
            'Configuration value not found, section={}, name={}'
            .format(section, name)
        )
    elif default is not None and value is no_value:
        value = default

    return value

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2023 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import typing as tp

from n23 import TaskResult

T = tp.TypeVar('T')
S = tp.TypeVar('S')

def fmap(f: tp.Callable[[T], S]) \
    -> tp.Callable[[TaskResult[T]], TaskResult[S]]: ...

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2023 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ruff: noqa: SLF001

"""
Tests for timer with timer tick.
"""

import time
from n23.timer import Timer

import pytest

@pytest.mark.asyncio
async def test_timer_tick_and_expirations() -> None:
    """
    Test timer initialization, expiration and tick count.
    """
    timer = Timer(0.1)
    timer.start()
    num_exp = await timer

    assert 1 == num_exp
    assert 0 == timer._tick

@pytest.mark.asyncio
async def test_timer_tick_and_expirations_over() -> None:
    """
    Test timer initialization, expiration and tick count.
    """
    timer = Timer(0.1)
    timer.start()

    num_exp = await timer
    num_exp = await timer
    num_exp = await timer

    time.sleep(0.2)
    num_exp = await timer

    assert 2 == num_exp

    # due to two expirations
    assert 4 == timer._tick

# vim: sw=4:et:ai



#
# n23 - data acquisition and processing framework
#
# Copyright (C) 2013-2023 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Timer tracking its own tick.
"""

import atimer
import logging
import typing as tp

logger = logging.getLogger(__name__)

P = tp.ParamSpec('P')

class Timer(atimer.Timer):
    """
    Timer tracking its own tick.
    """
    def __init__(self, interval: float) -> None:
        """
        Create new timer instance.

        .. seealso:: `atimer.Timer`
        .. seealso:: `atimer.Timer.start`
        """
        super().__init__(interval)
        self._tick = -1

    def __await__(self) -> tp.Generator[tp.Any, None, int]:
        """
        Sleep for time specified by timer interval.

        The coroutine returns number of expirations.

        .. seealso:: `atimer.Timer.start`
        """
        # sleep and count number of expirations
        num_exp = yield from super().__await__()

        # use number of expirations to update the timer tick
        self._tick += num_exp

        if num_exp > 1:
            logger.warning(
                'Timer expired more than one time: {}'.format(num_exp)
            )
        return num_exp

# vim: sw=4:et:ai
